/********************************************************************************
 *                                                                              *
 *                  Container.java written by Andrew Maunder                    *
 *                     Department of Computer Science                           *
 *                        University of Cape Town                               *
 *                                Aug 2004                                      *
 *******************************************************************************/

import java.net.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import java.text.*;

public class Container {

    private PrintWriter errorLog; // the error log
    private PrintWriter successLog; // the success log
    private ResponseThread responseThread; // handles to the two monitor threads
    private Pool threadPool;
    private Pool responsePool;
    private Hashtable servletList;
    private BufferedReader br;
    private static final int STARTTHREADS = 10;
    private static final int STARTTHREADS_RESPONSE = 10;
    private Date currentTime;

    public Container ( ){

        currentTime = new Date(); // create a new date object
        servletList = new Hashtable();

        try {
	          br = new BufferedReader ( new InputStreamReader ( System.in ));
	          errorLog = new PrintWriter( new FileWriter( "error.log" ));
              successLog = new PrintWriter( new FileWriter( "success.log" ));
            }catch ( Exception e )
            {
               e.printStackTrace();
            }

         initContainer(); // initialise the container

    }//end of default constructor

    public boolean initContainer ( ){

      /* We have to initialize the container to load its existing state */
      boolean shutdown = false;
      boolean existing = false;

      logSuccess(" Initializing container " );

      /* Initialise the thread pool, we will use our queue to model this */
      threadPool = new Pool( STARTTHREADS ); // create a new thread pool of default size (10)
      responsePool = new Pool ( STARTTHREADS_RESPONSE );

      /* Main thread starts monitoring the request queue and listens to messages */
      String msg = "";
      
      System.out.println ("@");

      while ( !shutdown )
         {
           /* The main thread monitors the state of the container as well as container messages */
           msg = checkRequests();

	       logSuccess("Got request at : " + System.currentTimeMillis()); //The time the system got a request

           if ( !(msg.equals("")) && !(msg.equals("shutdown")) ){
            processServletRequest(msg); // process any requests on the queue
           }else if ( msg.equals("shutdown") ) {
            shutdown = true;
           }

         }// main thread while loop

         System.err.println("About to shutdown");
         shutdownContainer(); // shutdown the container

      return true;
    }//end of method to init container

    public boolean shutdownContainer( ){

     /* We have to shutdown the container correctly when attempting to kill the JVM */
     logSuccess(" Shutting down container");
     logError(" Non-errorneous shutdown completed ");

     try {
         errorLog.close();
         successLog.close(); //close the log files
         threadPool.finalize(); //destroy the thread pool
         responsePool.finalize();
         System.err.println("Killed all thread pools");

     }catch( Exception e )
     {
      e.printStackTrace();
     }

    return true;

    }//end of method to shutdown the container

    public String checkRequests( ){

     String result = "";
     /* Create a reference to the new socket with the connection */
     //byte[] buffer = new byte[500];

     try {
          /* If the container is still alive continue processing input requests */
          //System.in.read(buffer); //read from standard in
	 result = br.readLine();

	 //result = new String(buffer).trim();

     } catch (Exception e) {
        System.err.println("Error reading request");
     }

     return result;
    }// end of method to check if container is busy

    public void updateHashTable( ){

       servletList.clear(); //temporarily clear the hash table

    }//end of method to update the hashtable

    public String preLoadServlet( String uri ){

       URLClassLoader loader = null;
       URL[] urls = null;
       URLStreamHandler streamHandler = null;  //the variables we need to load a class
       File classPath = null;
       Class myClass = null;
       String repository = "";

       if ( uri.indexOf('?') != -1 )
          uri = uri.substring(0,uri.indexOf('?')); //extract the parameters from the uri

       String identifier = uri.substring(1, uri.lastIndexOf("!"));
       String servletName = uri.substring(uri.lastIndexOf("/") + 1);
       String path = uri.substring( uri.lastIndexOf("!") + 1, uri.lastIndexOf("/"));

       //System.out.println( identifier + " " + servletName + " " + path );

       TableElement hashOut = (TableElement)servletList.get(servletName);

       /* Get the file modification details */
       String fileName = path + File.separator + servletName + ".class";
       File test = new File(fileName);

       if ( hashOut != null ){

         if ( test.lastModified() > hashOut.getLastMod() ){

         /* If an servlet class exists in the hashtable then check if it has been modified */
        try {
	         //create a URLClassLoader
             urls = new URL[1];
             streamHandler = null;
             classPath = new File(path); //this must find the users servlet directory
             repository = (new URL("file", null, classPath.getCanonicalPath() + File.separator )).toString() ;
             urls[0] = new URL(null, repository, streamHandler);
             loader = new URLClassLoader(urls);
             myClass = loader.loadClass(servletName); // dynamically load the class

             if ( servletList.size() < 100 )
                servletList.put(servletName,new TableElement(myClass, test.lastModified()) ); //add the serlvet class to hash table
             else {
                updateHashTable(); //clear the hashtable
             }

           }
           catch (Exception e)
           {
            System.out.println( "Error 505 : Servlet class not found." );
            }

         }else {
           myClass = hashOut.getServletClass();  //use the existing entry
           //System.err.println("Reuse in progress");
         }

       }else if ( hashOut == null ){

       /*If the class is not already loaded then dynamically load it*/
       try {
	   //System.err.println("Orig load");
        //create a URLClassLoader
        urls = new URL[1];
        streamHandler = null;
        classPath = new File(path); //this must find the users servlet directory
        repository = (new URL("file", null, classPath.getCanonicalPath() + File.separator )).toString() ;
        //System.err.println (repository);
        //repository = repository + "t2.class";
        //System.err.println (repository);
        urls[0] = new URL(null, repository, streamHandler);
        loader = new URLClassLoader(urls);

    }
    catch (IOException e)
    {
      System.out.println( "X1Error 505 : Servlet class not found." );
    }

     try {
           //System.err.println (servletName);
           myClass = loader.loadClass(servletName); // dynamically load the class
      }
      catch (ClassNotFoundException e) {
           System.out.println("X2Error 505 : Servlet class not found.");
      }

      servletList.put(servletName,new TableElement(myClass, test.lastModified()) ); //add the serlvet class to hash table
      }

     return servletName;
    }//end of preload servlet method

    public boolean processServletRequest ( String msg ){

    /* This method is run by the main thread and services a request from the queue */
    boolean complete = false;
    Class servletClass = null;
    /* Process the oldest element off the queue */

    try {
         /* Create a new servlet thread type with message data */
         String servletName = preLoadServlet(msg);
         TableElement hashOut = (TableElement)servletList.get(servletName);
         String id = msg.substring(1, msg.lastIndexOf("!"));

         /* Create the response monitor thread */
         PipedInputStream pin = new PipedInputStream();
         PipedOutputStream pout = new PipedOutputStream(pin);
         DataOutputStream out = new DataOutputStream( pout );

         ResponseThread rt = new ResponseThread( id, out, pin, pout );
         responsePool.assign(rt);
         responsePool.complete();

         PoolServletThread pst = new PoolServletThread( msg, hashOut.getServletClass(), out ); //create a new pool thread
         threadPool.assign(pst); // assign the job to the pool
         threadPool.complete(); // complete the task

         complete = true;
         }
    catch (Exception e)
    {
	//logError( " Error creating a new thread to process a request in pool." );
      complete = false;
    }
     //logSuccess( " Request processed from the pool." );

     return complete;

    }//end of request processing method

    public void logSuccess( String line ){

      DateFormat df = DateFormat.getDateTimeInstance(DateFormat.LONG, DateFormat.LONG);
      String time = df.format(currentTime);

      /* This method will write container events to the log file */
      successLog.println( time + " | " + line);

    }//end of method to log container successes

    public void logError( String line ){

      DateFormat df = DateFormat.getDateTimeInstance(DateFormat.LONG, DateFormat.LONG);
      String time = df.format(currentTime);

      /* This method will write an information line to the error log file */
      errorLog.println( time + " | " + line );

    }//end of method to log container errors

}//end of the container class