/* file: mod_x.c */

#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <http_protocol.h>
#include <ap_alloc.h>
#include <util_script.h>
/* Forward declaration so that ap_get_module_config() can find us. */

#include "config.h"

#define MAXDATASIZE 100 // max number of bytes we can get at once

module x_module;

/* Here's our per-directory configuration data */

typedef struct {
char *to;

} x_dir_config;

typedef struct {

int init_sockfd;
int counter;
} x_serv_config;

x_serv_config *serv_cfg; // handle to a server config

/* This function is called to create the default per-directory
configuration */

static void x_init(server_rec *s, pool *p)
{
#ifdef DEBUG
   fprintf(stderr, "mod_x: we are in the method init method\n");
#endif
   //possibly start Reinhardt's module

} // end of method initialisation method

static void *x_create_dir_config(pool *p, char *path)
{
   x_dir_config *cfg =
    (x_dir_config *)ap_pcalloc(p, sizeof(x_dir_config));
   cfg->to = "world";
   return (void *)cfg;
}

static int php_handler(request_rec *r) 
{
#ifdef DEBUG
   fprintf(stderr, "mod_x: we have called the php handler\n");
#endif

   serv_cfg = (x_serv_config *) ap_pcalloc(r->pool, sizeof(x_serv_config));

   // we will now set the file descriptor of the socket

   int sockfd, numbytes;
   char buf[MAXDATASIZE];
   struct hostent *he;
   struct sockaddr_in their_addr; // connector's address information

   if ((he=gethostbyname("localhost")) == NULL) 
   { // get the host info
      fprintf(stderr,"mod_x: ERROR: cannot find host\n");
   }

   if (( serv_cfg->init_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) 
      fprintf(stderr,"mod_x: ERROR: cannot connect to socket\n");

   their_addr.sin_family = AF_INET; // host byte order
   their_addr.sin_port = htons(PORT); // short, network byte order
   their_addr.sin_addr = *((struct in_addr *)he->h_addr);
   memset(&(their_addr.sin_zero), '\0', 8); // zero the rest of the struct

#ifdef DEBUG
   fprintf(stderr, "mod_x: about to connect to switch module\n");
   fprintf(stderr, "%i\n",serv_cfg->init_sockfd);
#endif

   if (connect(serv_cfg->init_sockfd, (struct sockaddr *)&their_addr, sizeof(struct sockaddr)) == -1) 
      fprintf(stderr,"mod_x: ERROR: cannot connect to host\n");

   if (send(serv_cfg->init_sockfd, r->unparsed_uri, strlen(r->unparsed_uri), 0) == -1)
      fprintf(stderr, "mod_x: ERROR: error sending data\n");

#ifdef DEBUG
   fprintf (stderr, "mod_x: URI %s LENGTH %d\n", r->unparsed_uri, strlen(r->unparsed_uri));
#endif

   //Receive the reply from the X-Switch module, block until then
   //if ( numbytes=recv(serv_cfg->init_sockfd, buf, MAXDATASIZE-1, 0) == -1 ){
   //fprintf(stderr,"Error receiving data\n");
   //}
   numbytes=recv(serv_cfg->init_sockfd, buf, MAXDATASIZE-1, 0);
   while (numbytes > 0)
   {
      buf[numbytes] = 0;
#ifdef DEBUG
      fprintf (stderr, "mod_x: received: %s\n", buf);
#endif
      ap_rprintf(r, "%s", buf);
      //memset (buf, 0, MAXDATASIZE);
      numbytes=recv(serv_cfg->init_sockfd, buf, MAXDATASIZE-1, 0);
   }

#ifdef DEBUG
   //fprintf (stderr, "Received: %s", buf);
   fprintf(stderr, "mod_x: out of receive loop in php handler\n");
   //ap_rprintf(r, "%s\n", buf);
   //ap_rprintf(r, "%s\n", r->unparsed_uri );
#endif
   
   return OK;
}

static int perl_handler(request_rec *r) 
{
#ifdef DEBUG
   fprintf(stderr, "mod_x: we have called the perl handler\n");
#endif

   serv_cfg = (x_serv_config *) ap_pcalloc(r->pool, sizeof(x_serv_config));

   // we will now set the file descriptor of the socket

   int sockfd, numbytes;
   char buf[MAXDATASIZE];
   struct hostent *he;
   struct sockaddr_in their_addr; // connector's address information

   if ((he=gethostbyname("localhost")) == NULL) 
   { // get the host info
      fprintf(stderr,"mod_x: ERROR: cannot find host\n");
   }

   if (( serv_cfg->init_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) 
      fprintf(stderr,"mod_x: ERROR: cannot connect to socket\n");

   their_addr.sin_family = AF_INET; // host byte order
   their_addr.sin_port = htons(PORT); // short, network byte order
   their_addr.sin_addr = *((struct in_addr *)he->h_addr);
   memset(&(their_addr.sin_zero), '\0', 8); // zero the rest of the struct

#ifdef DEBUG
   fprintf(stderr, "mod_x: about to connect to switch module\n");
   fprintf(stderr, "%i\n",serv_cfg->init_sockfd);
#endif

   if (connect(serv_cfg->init_sockfd, (struct sockaddr *)&their_addr, sizeof(struct sockaddr)) == -1) 
      fprintf(stderr,"mod_x: ERROR: cannot connect to host\n");

   if (send(serv_cfg->init_sockfd, r->unparsed_uri, strlen(r->unparsed_uri), 0) == -1)
      fprintf(stderr, "mod_x: ERROR: error sending data\n");

#ifdef DEBUG
   fprintf (stderr, "mod_x: URI %s LENGTH %d\n", r->unparsed_uri, strlen(r->unparsed_uri));
#endif

   //Receive the reply from the X-Switch module, block until then
   //if ( numbytes=recv(serv_cfg->init_sockfd, buf, MAXDATASIZE-1, 0) == -1 ){
   //fprintf(stderr,"Error receiving data\n");
   //}
   numbytes=recv(serv_cfg->init_sockfd, buf, MAXDATASIZE-1, 0);
   while (numbytes > 0)
   {
      buf[numbytes] = 0;
#ifdef DEBUG
      fprintf (stderr, "mod_x: received: %s\n", buf);
#endif
      ap_rprintf(r, "%s", buf);
      //memset (buf, 0, MAXDATASIZE);
      numbytes=recv(serv_cfg->init_sockfd, buf, MAXDATASIZE-1, 0);
   }

#ifdef DEBUG
   //fprintf (stderr, "Received: %s", buf);
   fprintf(stderr, "mod_x: out of receive loop in perl handler\n");
   //ap_rprintf(r, "%s\n", buf);
   //ap_rprintf(r, "%s\n", r->unparsed_uri );
#endif
   
   return OK;
}

static int servlet_handler(request_rec *r)
{
#ifdef DEBUG
   fprintf(stderr,"mod_x: we have called the servlet handler\n");
#endif

   serv_cfg = (x_serv_config *) ap_pcalloc(r->pool, sizeof(x_serv_config));

   // we will now set the file descriptor of the socket

   int sockfd, numbytes;
   char buf[MAXDATASIZE];
   struct hostent *he;
   struct sockaddr_in their_addr; // connector's address information

   if ((he=gethostbyname("localhost")) == NULL) 
   { // get the host info
      fprintf(stderr,"mod_x: ERROR: cannot find host\n");
   }

   if (( serv_cfg->init_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
      fprintf(stderr,"mod_x: ERROR: cannot connect to socket\n");

   their_addr.sin_family = AF_INET; // host byte order
   their_addr.sin_port = htons(PORT); // short, network byte order
   their_addr.sin_addr = *((struct in_addr *)he->h_addr);
   memset(&(their_addr.sin_zero), '\0', 8); // zero the rest of the struct

#ifdef DEBUG
   fprintf(stderr, "mod_x: about to connect to switch module\n");
   fprintf(stderr, "%i\n",serv_cfg->init_sockfd);
#endif

   if (connect(serv_cfg->init_sockfd, (struct sockaddr *)&their_addr, sizeof(struct sockaddr)) == -1)
      fprintf(stderr,"mod_x: ERROR: cannot connect to host\n");

   if (send(serv_cfg->init_sockfd, r->unparsed_uri, 99, 0) == -1)
      fprintf(stderr, "mod_x: ERROR: error sending data\n");

   // Receive the reply from the X-Switch module, block until then
   //if ( numbytes=recv(serv_cfg->init_sockfd, buf, MAXDATASIZE-1, 0) == -1 ){
   //fprintf(stderr,"Error receiving data\n");
   //}
   numbytes = recv(serv_cfg->init_sockfd, buf, MAXDATASIZE-1, 0);
   while (numbytes > 0)
   {
      buf[numbytes] = 0;
#ifdef DEBUG
      fprintf (stderr, "mod_x: received: %s\n", buf);
#endif
      ap_rprintf(r, "%s", buf);
      //memset (buf, 0, MAXDATASIZE);
      numbytes = recv(serv_cfg->init_sockfd, buf, MAXDATASIZE-1, 0);
   }

#ifdef DEBUG
   fprintf(stderr, "mod_x: out of receive loop in servlet handler\n");
   //ap_rprintf(r, "%s\n", buf);
   //ap_rprintf(r, "%s\n", r->unparsed_uri );
#endif

   return OK;
}

/* Make the name of the content handler known to Apache */

static handler_rec x_handlers[] =
{
   {"servlet-handler", servlet_handler},
   {"php-handler", php_handler},
   {"perl-handler", perl_handler},
   {NULL}
};

/* Tell Apache what phases of the transaction we handle */

module MODULE_VAR_EXPORT x_module =
{
   STANDARD_MODULE_STUFF,
   x_init, /* module initializer */
   x_create_dir_config, /* per-directory config creator */
   NULL, /* dir config merger */
   NULL, /* server config creator */
   NULL, /* server config merger */
   NULL, /* command table */
   x_handlers, /* [9] content handlers */
   NULL, /* [2] URI-to-filename translation */
   NULL, /* [5] check/validate user_id */
   NULL, /* [6] check user_id is valid *here* */
   NULL, /* [4] check access by host address */
   NULL, /* [7] MIME type checker/setter */
   NULL, /* [8] fixups */
   NULL, /* [10] logger */
   NULL, /* [3] header parser */
   NULL, /* process initialization */
   NULL, /* process exit/cleanup */
   NULL /* [1] post read_request handling */
};

