/***************************************************************************
*   Copyright (C) 2004 by Reinhardt van Rooyen                             *
*   rvanrooy@cs.uct.ac.za						   *
*  									   *
*   The xSwitch program is the intellectual property of Andrew Maunder     *
*   and Reinhardt van Rooyen, along with with computer science department  *  
*   at the University of Cape Town                                         *
*									   *
*   This is part of the xSwitch honours project by the authors             *
*                                                                          *
***************************************************************************/

#include <time.h>

#include "switch.h"
#include "../mod_x/config.h"

#ifdef DEBUG
#define VERBOSE
#endif

#define FORCOUNT 300
#define ERROR_LOG1

#define MAX_USER_ENGINE_NUM 3
#define NUM_USERS 10
#define ARRAY_SIZE 11

int initialiseEngine(Engine * ptr);
int findUserEngine(Engine * ptr, int hash,char * owner);
int findAvailableEngine(Engine * ptr, int hash);
Engine createNewEngine(char * owner,char * type,char * request,int socket);
void sendRequest(Engine * ptr,char * type,char * request,int socket);

void sigchld_handler(int s);


struct UserInfo createUserInfoEntry(char * owner,char * type);
void decreaseNumRequests(char * owner,char * type);
void increaseNumRequests(char * owner,char * type);

//void sighup();
struct InfoEngine
{
	//communication between processes
	int scktpair[2];
	int tempScktpair[2];
	//variables set for receiving response from child
	
	//thread for engine output
	pthread_t engineThread;
	int  threadCreateValue;
	
	//pthread_t engineThreadControl;
	//int  controlThreadCreateValue;
	
	// synchronisation flags for engine startup
	pthread_mutex_t condition_mutex;
	pthread_cond_t condition_cond;
	
	//ID of engine as incremental int
	int id;
	//type of engine
	char type[ENGINE_TYPE_LENGTH];
	//the status of engine
	int status;
	//the user who owns the process
	char owner[USER_LENGTH];
	//the uid of the owner who owns the process
	int uid;
	//the path to the users directory where the files are stored
	char path[MAX_LENGTH];
};

struct UserInfo
{
	char owner[USER_LENGTH];
	char type[1024];
	int numEngines;
	int totalRequests;
	int currentRequests;
	struct InfoEngine *ptrArray[MAX_USER_ENGINE_NUM];
	int infoscktpair[2];
	int uid;
	//the path to the users directory where the files are stored
	char path[MAX_LENGTH];
};

int findAvailablePlace(struct UserInfo * ptr, int hash);
int findUserInfo(struct UserInfo * ptr, int hash,char * owner);
struct InfoEngine * createNewInfoEngine(char * owner,char * type,int fd,int fd2);
void infosendRequest(struct UserInfo * ptr,char * type,char * request,int socket);
int new_fd;
struct UserInfo userInfo[ARRAY_SIZE];
//FILE * errorLog;
//FILE * log;

int main()
{

  printf("Program started");
//Set up the logs,
#ifdef ERROR_LOG
errorLog = fopen ("error_log","a");
fprintf(errorLog,"_____----- PROGRAM STARTED -----_____\n");

#endif

#ifdef VERBOSE
printf("_____----- PROGRAM STARTED -----_____\n");
#endif



//gEngine engineArray[ARRAY_SIZE];


memset(&userInfo,0,sizeof(userInfo));
//memset(&engineArray,0,sizeof(engineArray));
int hash = 1;

#ifdef VERBOSE
printf("MAIN: #%d, Starting program\n",__LINE__);
#endif
	
	int mainProgram = CONTINUE;
	
	int sockfd;  // listen on sock_fd, new connection on new_fd
	struct sockaddr_in my_addr;    // my address information
	struct sockaddr_in their_addr; // connector's address information
	int sin_size;
	struct sigaction sa;
	int yes=1;
	
	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket");
		exit(1);
	}
	
	if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
		perror("setsockopt");
		exit(1);
	}
	
	my_addr.sin_family = AF_INET;         // host byte order
	my_addr.sin_port = htons(PORT);     // short, network byte order
	my_addr.sin_addr.s_addr = INADDR_ANY; // automatically fill with my IP
	memset(&(my_addr.sin_zero), '\0', 8); // zero the rest of the struct
	
	if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) {
		perror("bind");
		exit(1);
	}
	
	if (listen(sockfd, BACKLOG) == -1) {
		perror("listen");
		exit(1);
	}
	
	sa.sa_handler = sigchld_handler; // reap all dead processes
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa, NULL) == -1) {
		perror("sigaction");
		exit(1);
	}

	while (mainProgram == CONTINUE)
	{
		sin_size = sizeof(struct sockaddr_in);
		if ((new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size)) == -1) {
			perror("accept");
			continue;
		}
#ifdef VERBOSE
printf("MAIN: #%d, Got connection\n",__LINE__);
#endif		
		//printf("server: got connection from %s\n",inet_ntoa(their_addr.sin_addr));

		int lengthOfInput;
		char inputLine[1028];
            if ((lengthOfInput=recv(new_fd, inputLine, 1027, 0)) == -1)
	    {
              fprintf(stderr,"Error receiving data\n");
#ifdef ERROR_LOG
fprintf(errorLog,"MAIN: LINE %d:, Error Receiving data from socket\n",__LINE__);
#endif			   
	  }
	   
	   if (lengthOfInput > 3) {
#ifdef VERBOSE              
//fprintf(stdout, "%s LENGTH: %d\n", inputLine, strlen(inputLine) );
#endif		

			//make sure request isn't bogus		
			if (lengthOfInput < 3)
			{

			}
#ifdef VERBOSE
printf("MAIN: #%d, Got Request\n",__LINE__);
#endif
			inputLine[lengthOfInput] = 0;
			
			//split input into manageable bits, should probably be a method
			char * owner;
			char * type;
			char * request;
			//evaluate request,
			
			//request URI looks like
			// /~reinhardt/php/index.php?name=Billy
			// /~reinhardt/servlet/bob?name=Bob
printf ("BEFORE TOKENISATION %s of length %d\n", inputLine, strlen(inputLine));
			
			owner = (char *)strtok(inputLine,"~/");
			type = (char *)strtok(NULL,"/");
			request = (char *)strtok(NULL,"/");
printf ("AFTER TOKENISATION %s of length %d\n", request, strlen(request));
			
			int hash = computeHash(owner,type);
#ifdef VERBOSE
printf("MAIN: #%d, Check if User has engine...\n",__LINE__);
#endif

			//need to check engineArray for existing engine
			
			int j = 0;
//			printf("-------------- %s and hash = %d\n",engineArray[hash].owner,hash);
			if ((userInfo[hash].owner)&&(j = (strcmp(userInfo[hash].owner,"")) != 0))
			{
				
				//something exists in hash posistion
				//need to find out whether the particular engine belongs to user.
				
				//int k = findUserEngine(&userInfo[hash],hash,owner);
				int k = findUserInfo(&userInfo[hash],hash,owner);
				//printf("k = %d\n",k);
				if (k == -1)
				{
#ifdef VERBOSE
printf("MAIN: #%d: No engine was found for user, but hash already exists. find a place to create one\n",__LINE__);
#endif
					int newhash = findAvailablePlace(&userInfo[hash],hash);
					

#ifdef VERBOSE
printf("MAIN: #%d, No, user needs one at hash posisiton %d (orig hash = %d)\n",__LINE__,newhash,hash);
#endif					
					hash = newhash;
					
					
			
			struct InfoEngine *tempE = createNewInfoEngine(userInfo[hash].owner,userInfo[hash].type,userInfo[hash].infoscktpair[1],userInfo[hash].infoscktpair[0]);
				userInfo[hash].ptrArray[0] = tempE;
				
				
#ifdef VERBOSE
printf("MAIN: #%d: new_fd = %d\n",__LINE__,new_fd);
#endif				
				//infosendRequest(&userInfo[hash],userInfo[hash].type,request,new_fd);
//				printf("*%d\n",userInfo[hash].currentRequests);
/*	if (userInfo[hash].currentRequests > 2*userInfo[hash].numEngines)
				{
					printf("\n-----------------------\nANOTHER ENGINE CREATED  for user %s\n---------------\n",userInfo[hash].owner);
					//check to see if the user has reached his maximum amount of engines.
					
					if (userInfo[hash].numEngines < MAX_USER_ENGINE_NUM)
					{
					
						//create another engine for the user
						struct InfoEngine tempE = createNewInfoEngine(userInfo[hash].owner,userInfo[hash].type,userInfo[hash].infoscktpair[1],userInfo[hash].infoscktpair[0]);
						userInfo[hash].ptrArray[userInfo[hash].numEngines] = &tempE;
						userInfo[hash].numEngines++;
						int l = 0;
						for (l = 0;l < FORCOUNT;l++)
						{
						}
					}
					//infosendRequest(&userInfo[hash],userInfo[hash].type,request,new_fd);
				
				}/**/
				infosendRequest(&userInfo[hash],userInfo[hash].type,request,new_fd);
				}
				else
				{	
#ifdef VERBOSE
printf("MAIN: #%d, Yes, user \"%s\" has one at hash posisiton %d\n",__LINE__,userInfo[k].owner,hash);
#endif					

					hash = k;
					
#ifdef VERBOSE
printf("MAIN: #%d: new_fd = %d\n",__LINE__,new_fd);
#endif				
//printf("*%d\n",userInfo[hash].currentRequests);
/*				if (userInfo[hash].currentRequests > 2*userInfo[hash].numEngines)
					{
				//	printf("\n----------------------\n!ANOTHER ENGINE CREATED  for user %s\n---------------\n",userInfo[hash].owner);
						//create another engine for the user
						if (userInfo[hash].numEngines < MAX_USER_ENGINE_NUM)
						{
							struct InfoEngine tempE = createNewInfoEngine(userInfo[hash].owner,userInfo[hash].type,userInfo[hash].infoscktpair[1],userInfo[hash].infoscktpair[0]);
						//	printf("engine created sucessfully, adding it to the array of engines for user\n");
							userInfo[hash].ptrArray[userInfo[hash].numEngines] = &tempE;
							userInfo[hash].numEngines++;
						//	printf("letting it sleep a little while\n");
							int l = 0;
							for (l = 0;l < FORCOUNT;l++)
							{
							}
						//	printf("sending the request to the sockets\n");
							infosendRequest(&userInfo[hash],userInfo[hash].type,request,new_fd);
						}
					}/**/
					infosendRequest(&userInfo[hash],userInfo[hash].type,request,new_fd);
				}
				
			} // end of check to see if there is an engine at the hash
			else
			{
#ifdef VERBOSE
printf("MAIN: #%d, No, user \"%s\" does not have one. Hash posisiton = %d\n",__LINE__,owner,hash);
#endif					

			//	printf("NO ENGINES FOUND\n");
				userInfo[hash] = createUserInfoEntry(owner,type);
				
			struct InfoEngine *tempE = createNewInfoEngine(userInfo[hash].owner,userInfo[hash].type,userInfo[hash].infoscktpair[1],userInfo[hash].infoscktpair[0]);
				userInfo[hash].ptrArray[0] = tempE;
				userInfo[hash].numEngines++;
				//sleep(1);
				//int l = 0;
				//	for (l = 0;l < FORCOUNT;l++)
				//	{
				//	}
#ifdef VERBOSE
printf("MAIN: #%d: new_fd = %d\n",__LINE__,new_fd);
#endif		
printf("current Requests = %d\n__________________________________________\n",userInfo[hash].currentRequests);
userInfo[hash].currentRequests = 0;
				//infosendRequest(&userInfo[hash],userInfo[hash].type,request,new_fd);
 				//sleep(1);
#ifdef VERBOSE
printf ("MAIN: #%d: Engine status = %d\n", __LINE__, tempE->status);
#endif
	// wait for signal from engine's handling thread before sending request
				pthread_mutex_lock (&(tempE->condition_mutex));
				pthread_cond_wait (&(tempE->condition_cond), &(tempE->condition_mutex));
				pthread_mutex_unlock (&(tempE->condition_mutex));
#ifdef VERBOSE
printf ("MAIN: #%d: Engine status = %d\n", __LINE__, tempE->status);
#endif
				infosendRequest(&userInfo[hash],userInfo[hash].type,request,new_fd);
				
			}
		}
		else
		{
		printf("Invalid input\n");
		
#ifdef ERROR_LOG
fprintf(errorLog,"MAIN: LINE %d: Data received is invalid (too short, only 2 or less chars\n",__LINE__);
#endif
			//an error occurred reading from the input stream
			mainProgram = STOP;
		}// end of 

	}// end while loop to keep module persisten

	return 1;
} // end of the main method
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void *thread_control( void *ptr )
{
/*	int currentCount = 0;
	int alive =1;
	Engine *e;
	e = (Engine *) ptr;
	sleep(TIMEOUT_VALUE);
	while (alive == 1)
	{
		if ((currentCount == e->counter) && (e->status == STATUS_BUSY))
		{
			printf("TIMER HAS NOT MOVED ON, SCRIPT HAS TIMED OUT !!!\n");
			printf("killing child process now!\n");
			kill(e->id,SIGKILL);
			alive = 0;
		}
		else
		{
		
		}
		sleep(TIMEOUT_VALUE);
	}*/
}// end of thread_control
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int getUserUID(char * name)
{
	struct passwd *pw;
	int *uid;
	if (!(pw = getpwnam(name)))
	{
#ifdef ERROR_LOG
fprintf(errorLog,"getUserPath: LINE %d: Could not resolve username %s\n",__LINE__,name);
#endif	
		printf("Could not resolve username\n");
	}
	else
	{
	  uid = &(pw->pw_uid);
	} 
	return *uid;
} // end of getUserID
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
char * getUserPath(char * name)
{
//	printf("USER = %s\n",name);
	struct passwd *pw;
	char *path;
	path = (char *)malloc(MAX_LENGTH);
//	printf("going to resolve name\n");
	if (!(pw = getpwnam(name)))
	{
#ifdef ERROR_LOG
fprintf(errorLog,"getUserPath: LINE %d: Could not resolve username %s\n",__LINE__,name);
#endif
		printf("Could not resolve username\n");
	}
	else
	{

		char * dir;
		path = (pw->pw_dir);

	} 
	strcat(path,"/public_html/");
	return path;

} // end of getUserPath
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int computeHash(char * name, char * type)
{
  int total = 0;
	int i = 0;
	int max = strlen(name);
	for (i; i < max;i++)
	{
	total += name[i];
	}
	 
	max = strlen(type);
	i = 0;
	for (i; i < max;i++)
	{
	total += type[i];
	}
#ifdef VERBOSE
printf("computeHash: LINE %d: hash computed for user %s = %d\n",__LINE__,name,total%11);
#endif
	return total % 11;
	
	
}// end of compute hash
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int findUserInfo(struct UserInfo *ptr, int hash,char *owner)
{
	int cnt = hash;
	int i = 0;
	int j = 0;
	while(cnt < ARRAY_SIZE)
	{
		if (strcmp(ptr->owner,"") == 0)
		{
			return -1;
		}
		if (i = (strcmp(ptr->owner,owner)) == 0)
		{
			return hash+j;
		}
		ptr = ptr+1;
		j++;
	}//end of while loop

}// end of find user info (in array)
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int findUserEngine(Engine * ptr, int hash,char * owner)
{
int cnt = hash;
int i = 0;
int j = 0;
	while(cnt < ARRAY_SIZE)
	{
		if (strcmp(ptr->owner,"") == 0)
		{
			return -1;
			}
		if (i = (strcmp(ptr->owner,owner)) == 0)
		{
			return hash+j;
		}
	ptr = ptr+1;
	j++;
	}//end of while loop

}// end of findUserEngine
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int findAvailableEngine(Engine * ptr, int hash)
{
//find next empty or DELETED engine
//and maybe resize the array if no space
int go = 1;
int j = 0;
	while (go)
	{
		if ((ptr->owner)&&((j = (strcmp(ptr->owner,"")||(ptr->status == STATUS_DELETED) != 0))))
		{
			return hash +j;
		}
		else
		{
			ptr++;
			++hash%ARRAY_SIZE;
			j++;
		}
	}
} // end of find available engine
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void infosendRequest(struct UserInfo * ptr,char * type,char * request,int socket)
{
//printf("!!!!!!!!!!!!!!!!!! currentRequest cnt = %d\n",ptr->currentRequests);
	char pathrequest[OUTPUT_LENGTH];

	sprintf(pathrequest,"!%d!",socket);

	strcat(pathrequest,ptr->path);
	strcat(pathrequest,type);
	strcat(pathrequest,"/");
	strcat(pathrequest,request);
#ifdef VERBOSE
	printf ("LAST CHAR: %d\n", pathrequest[strlen(pathrequest)-1]);
#endif
	strcat(pathrequest,"\n");
	
#ifdef VERBOSE
	printf ("SEND request: %s", pathrequest);
#endif
	
	write(ptr->infoscktpair[0],pathrequest,strlen(pathrequest));
#ifdef VERBOSE
	printf ("PATH LENGTH: %d PATH %s\n", strlen(pathrequest), pathrequest);
#endif

}//end of infosendRequest
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void sigchld_handler(int s)
{
    while(wait(NULL) > 0);
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
struct UserInfo createUserInfoEntry(char * owner,char * type)
{
	struct UserInfo temp;
	//temp.owner = owner;
	//temp.type = type;
	strcpy (temp.owner, owner);
	strcpy (temp.type, type);
	temp.numEngines = 0;
	temp.currentRequests = 0;
	strcpy(temp.path,getUserPath(temp.owner));
	socketpair(AF_UNIX, SOCK_STREAM, 0, temp.infoscktpair);

	
	return  temp;
}// End of createUserInfoEntry
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
struct InfoEngine *createNewInfoEngine(char * owner,char * type,int fd,int fd2)
{


struct InfoEngine *temp;
temp = (struct InfoEngine *)malloc(sizeof (struct InfoEngine));
memset(temp->owner,0,USER_LENGTH);
memset(temp->type,0,ENGINE_TYPE_LENGTH);
strcpy(temp->owner,owner);
strcpy(temp->type,type);

int x = socketpair(AF_UNIX, SOCK_STREAM, 0, temp->scktpair);
int y = socketpair(AF_UNIX, SOCK_STREAM, 0, temp->tempScktpair);
#ifdef VERBOSE
   printf ("socketpair : %d %d\n", x, y);
#endif

strcpy(temp->path,getUserPath(temp->owner));

temp->uid = getUserUID(temp->owner);
 int tuid = temp->uid;
temp->status = STATUS_BUSY; // make sure engine is initially set as unavailable
                            // so no requests arrive before it has started up

#ifdef VERBOSE
printf("createNewInfoEngine LINE %d: Executing engine type\n",__LINE__,type);
#endif		

int fid = fork();
temp->id = fid;
	if (fid == 0)
	{
                char command[2048];
                char comm2[2048];
                if (strcmp(temp->type,"php") == 0)
                   strcpy (comm2, "php ../con_php/phpmod.php");
                else if (strcmp(temp->type,"servlet") == 0)
                   strcpy (comm2, "java -classpath ../con_java:/usr/local/share/java/classes/servlet-api.jar Driver");
                else if (strcmp(temp->type,"perl") == 0)
                   strcpy (comm2, "perl ../con_perl/perlmod.pl");
                sprintf (command, "%d %d %s", temp->uid, fd, comm2);
#ifdef VERBOSE
   printf ("COMMAND: %s\n", command);
#endif

                close (new_fd); // close socket from mod_x since the socket
                                // is duplicated on forking and this
                                // child process will morph into an engine
		
		close(0); // close standard input stream
		close(1); // close standard output stream
		
		if (dup (temp->tempScktpair[1]) != 0 || dup (temp->scktpair[1]) != 1)
		{
			perror ("error duplicating socket for stdin/stdout/stderr");
			exit (1);
		}// end of changing input and output streams.
		
		setbuf(stdout,NULL);

		execl("suexecme","suexecme",command,NULL);
		
	}// end of  child process instructions
	else
	{
		//parent process
		//printf("engine process id = %d\n",temp.id);
		//printf("owner = %s\n",temp.owner);
		
               //create the thread required to run the engine
		
#ifdef VERBOSE
		printf ("PARENT owner %s", temp->owner);
#endif
		temp->condition_mutex = PTHREAD_MUTEX_INITIALIZER;
		temp->condition_cond = PTHREAD_COND_INITIALIZER;
		temp->threadCreateValue =  pthread_create(&(temp->engineThread), NULL, info_listen_for_child, temp);

		return temp;
		
	}// end if parent instructions for fork	

   return temp;
} // end of Create new info engine
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void *info_listen_for_child( void *ptr )
{


//printf("start of child thread\n");/~reinhardt/php/index.php?name=sweet

   int fd;
   struct InfoEngine *e;
   char owner[USER_LENGTH];
   char type[ENGINE_TYPE_LENGTH];
   int id;
	
#ifdef VERBOSE
   printf("THREAD LINE %d: Converting void pointer to engine pointer.\n",__LINE__);
#endif     
   e = (struct InfoEngine *)ptr;
   strcpy(owner,e->owner);
   strcpy(type,e->type);
   id = e->id;

   char *line = (char *)malloc(OUTPUT_LENGTH);
	
   ssize_t xread;
#ifdef VERBOSE
   printf ("THREAD owner %s\n", owner);	
   printf ("THREAD socket %u\n", e->scktpair[0]);
#endif	
   while (threadExit == 0)
   {
	
#ifdef VERBOSE
      printf("THREAD start collecting a line\n");
#endif	
	
      char onechar;
      xread = 0;
      onechar = 0;
      while ((onechar != 10) && ((xread+1) < OUTPUT_LENGTH))
      {
         read (e->scktpair[0], &onechar, 1);
   	 line[xread++] = onechar;
      }
      line[xread] = 0;
	
#ifdef VERBOSE
      printf("THREAD LINE %d: getline =%s=\n",__LINE__,line);
#endif

      if(line[0] == '+')
      {
         char * socketDes;
	 socketDes = strtok(line,"+");
	 fd = atoi(socketDes);
#ifdef VERBOSE
         printf("THREAD LINE %d: write to FD %d %s\n",__LINE__,fd);
#endif		
      }
      else if (line[0] =='@')
      {
#ifdef VERBOSE
         printf("THREAD LINE %d: got control statement @\n",__LINE__);
#endif	
	 e->status = STATUS_READY;
	 // signal to parent process that engine is ready
				pthread_mutex_lock (&(e->condition_mutex));
				pthread_cond_signal (&(e->condition_cond));
				pthread_mutex_unlock (&(e->condition_mutex));
      }
      else if (xread == 2)
      {
         //control character found
	 if (line[0] == '$')
	 {
	    e->status = STATUS_READY;
#ifdef VERBOSE
            printf("THREAD LINE %d: about to close fd %d \n",__LINE__,fd);
#endif			
	 }
	 if (fd != 0)
	 {
	    decreaseNumRequests(owner,type);
	    int returnc = close(fd);
            fd = 0;
#ifdef VERBOSE
            printf("THREAD LINE %d: close fd returned with %d\n",__LINE__,returnc);
#endif			
	 }
#ifdef VERBOSE
         printf("THREAD LINE %d: Status set to READY for %s and fd %d closed\n",__LINE__,owner,fd);
#endif		
      }
      else
      {
#ifdef VERBOSE
         printf("THREAD LINE %d: Output from %s %s engine  (length = %d)-- %s\n",__LINE__,e->owner,e->type,strlen(line),line);
#endif
	 printf("engine id = %d\n",id);
	 //printf("!@#!@#!@# length of line = %d\n",strlen(line));
	 //printf("******** socket fd = %d  **********\n",fd);
			
#ifdef VERBOSE
         printf("THREAD LINE %d: Writing output to fd %d\n",__LINE__,fd);
#endif			
	 write(fd, line, strlen(line));
	 //printf ("%s", line);
      }
   }
} // end of info listen for child
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int findAvailablePlace(struct UserInfo *ptr, int hash)
{
int go = 1;
int j = 0;
	while (go)
	{
		if ((ptr->owner)&&(j = (strcmp(ptr->owner,""))))
		{
			return hash +j;
		}
		else
		{
			ptr++;
			++hash%ARRAY_SIZE;
			j++;
		}
	}

} //end of find available place (in array)
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void decreaseNumRequests(char * owner,char * type)
{
//return;
	int h = computeHash(owner,type);

	int k = findUserInfo(&userInfo[h],h,owner);
	int temp = userInfo[k].currentRequests;
//	printf("$$$$$$$$$$$$$$$$$$$$$$\n\nDECREASE CALLED current Requests = %d\n",temp);
	userInfo[k].currentRequests = temp - 1;
	
	printf("%d\n",userInfo[k].currentRequests);

}//end of decrease Num Requests
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void increaseNumRequests(char * owner,char * type)
{
//return;
	int h = computeHash(owner,type);

	int k = findUserInfo(&userInfo[h],h,owner);
	int temp = userInfo[k].currentRequests;
//printf("@@@@@@@@@@@@@@@@@@@@@ currentRequests = %d for owner %s\n",temp,owner);	
	userInfo[k].currentRequests = temp + 1;
	userInfo[k].totalRequests = userInfo[k].totalRequests + 1;
	printf("%d, total requests = %d\n",userInfo[k].currentRequests,userInfo[h].totalRequests);

}//end of increase Num Requests
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
